@@ -11,25 +11,28 @@
 //! - Progress events for fetches
 //!
 //! ```
+//! let (events_tx, events_rx) = tokio::sync::mpsc::unbounded_channel();
+//!
+//! let shutdown = async_shutdown::Shutdown::new();
+//!
 //! let results_stream = Fetcher::default()
 //!     // Define a max number of ranged connections per file.
 //!     .connections_per_file(4)
 //!     // Max size of a connection's part, concatenated on completion.
 //!     .max_part_size(4 * 1024 * 1024)
-//!     // An `Arc<AtomicBool>` which can be used to interrupt the download.
-//!     .cancel(cancellable)
 //!     // The channel for sending progress notifications.
 //!     .events(events_tx)
 //!     // Maximum number of retry attempts.
 //!     .retries(3)
+//!     // Cancels the fetching process when a shutdown is triggered.
+//!     .shutdown(shutdown)
 //!     // How long to wait before aborting a download that hasn't progressed.
 //!     .timeout(Duration::from_secs(15))
 //!     // Finalize the struct into an `Arc` for use with fetching.
 //!     .build()
 //!     // Take a stream of `Source` inputs and generate a stream of fetches.
-//!     .requests_stream(input_stream)
-//!     // Fetches up to 8 sources concurrently
-//!     .buffered(8);
+//!     // Spawns
+//!     .stream_from(input_stream, 4)
 //! ```
 
 #[macro_use]
@@ -41,7 +44,9 @@ extern crate log;
 #[macro_use]
 extern crate thiserror;
 
-pub mod checksum;
+pub mod iface;
+
+mod checksum;
 mod checksum_system;
 mod concatenator;
 mod get;
@@ -49,7 +54,9 @@ mod get_many;
 mod range;
 mod source;
 mod time;
+mod utils;
 
+pub use self::checksum::*;
 pub use self::checksum_system::*;
 pub use self::concatenator::*;
 pub use self::source::*;
@@ -57,40 +64,40 @@ pub use self::source::*;
 use self::get::{get, FetchLocation};
 use self::get_many::get_many;
 use self::time::{date_as_timestamp, update_modified};
-
+use async_shutdown::Shutdown;
 use futures::{
     prelude::*,
     stream::{self, StreamExt},
 };
 use http::StatusCode;
 use httpdate::HttpDate;
-use isahc::config::Configurable;
-use isahc::{AsyncBody, HttpClient as Client, Request, Response};
 use numtoa::NumToA;
+use reqwest::{Client, Response};
+use std::sync::atomic::Ordering;
 use std::{
     fmt::Debug,
-    future::Future,
     io,
     path::Path,
-    sync::{
-        atomic::{AtomicBool, Ordering},
-        Arc,
-    },
+    pin::Pin,
+    sync::{atomic::AtomicU16, Arc},
     time::{Duration, UNIX_EPOCH},
 };
 use tokio::fs;
 use tokio::sync::mpsc;
 
+/// The result of a fetched task from a stream of input sources.
+pub type AsyncFetchOutput<Data> = (Arc<Path>, Arc<Data>, Result<(), Error>);
+
+/// A channel for sending `FetchEvent`s to.
 pub type EventSender<Data> = mpsc::UnboundedSender<(Arc<Path>, Data, FetchEvent)>;
-pub type Output<T> = (Arc<Path>, Result<T, Error>);
 
 /// An error from the asynchronous file fetcher.
 #[derive(Debug, Error)]
 pub enum Error {
-    #[error("task was cancelled")]
-    Cancelled,
+    #[error("task was canceled")]
+    Canceled,
     #[error("http client error")]
-    Client(isahc::Error),
+    Client(#[source] reqwest::Error),
     #[error("unable to concatenate fetched parts")]
     Concatenate(#[source] io::Error),
     #[error("unable to create file")]
@@ -103,6 +110,8 @@ pub enum Error {
     MetadataRemove(#[source] io::Error),
     #[error("destination has no file name")]
     Nameless,
+    #[error("network connection was interrupted while fetching")]
+    NetworkChanged,
     #[error("unable to open fetched part")]
     OpenPart(Arc<Path>, #[source] io::Error),
     #[error("destination lacks parent")]
@@ -111,6 +120,8 @@ pub enum Error {
     TimedOut,
     #[error("error writing to file")]
     Write(#[source] io::Error),
+    #[error("fetch error")]
+    Read(#[source] reqwest::Error),
     #[error("failed to rename partial to destination")]
     Rename(#[source] io::Error),
     #[error("server responded with an error: {}", _0)]
@@ -119,8 +130,8 @@ pub enum Error {
     TokioSpawn(#[source] tokio::task::JoinError),
 }
 
-impl From<isahc::Error> for Error {
-    fn from(e: isahc::Error) -> Self {
+impl From<reqwest::Error> for Error {
+    fn from(e: reqwest::Error) -> Self {
         Self::Client(e)
     }
 }
@@ -138,10 +149,6 @@ pub enum FetchEvent {
     Fetching,
     /// Reports the amount of bytes that have been read for a file.
     Progress(u64),
-    /// Reports that a part of a file is being fetched.
-    PartFetching(u64),
-    /// Reports that a part has been fetched.
-    PartFetched(u64),
     /// Notification that a fetch is being re-attempted.
     Retrying,
 }
@@ -157,17 +164,18 @@ pub struct Fetcher<Data> {
     #[setters(skip)]
     client: Client,
 
-    /// When set, cancels any active operations.
-    #[new(default)]
-    #[setters(strip_option)]
-    cancel: Option<Arc<AtomicBool>>,
-
     /// The number of concurrent connections to sustain per file being fetched.
     /// # Note
     /// Defaults to 1 connection
     #[new(value = "1")]
     connections_per_file: u16,
 
+    /// Configure the delay between file requests.
+    /// # Note
+    /// Defaults to no delay
+    #[new(value = "0")]
+    delay_between_requests: u64,
+
     /// The number of attempts to make when a request fails.
     /// # Note
     /// Defaults to 3 retries.
@@ -180,6 +188,12 @@ pub struct Fetcher<Data> {
     #[new(value = "2 * 1024 * 1024")]
     max_part_size: u32,
 
+    /// Time in ms between progress messages
+    /// # Note
+    /// Defaults to 500.
+    #[new(value = "500")]
+    progress_interval: u64,
+
     /// The time to wait between chunks before giving up.
     #[new(default)]
     #[setters(strip_option)]
@@ -190,17 +204,20 @@ pub struct Fetcher<Data> {
     #[setters(into)]
     #[setters(strip_option)]
     events: Option<Arc<EventSender<Arc<Data>>>>,
+
+    /// Utilized to know when to shut down the fetching process.
+    #[new(value = "Shutdown::new()")]
+    shutdown: Shutdown,
 }
 
 impl<Data> Default for Fetcher<Data> {
     fn default() -> Self {
-        Self::new(
-            Client::builder()
-                .low_speed_timeout(1, std::time::Duration::from_secs(15))
-                .redirect_policy(isahc::config::RedirectPolicy::Follow)
-                .build()
-                .expect("failed to build HTTP client"),
-        )
+        let client = Client::builder()
+            .pool_idle_timeout(std::time::Duration::from_secs(20))
+            .build()
+            .unwrap();
+
+        Self::new(client)
     }
 }
 
@@ -210,41 +227,59 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
         Arc::new(self)
     }
 
-    /// Build a stream that will perform fetches when polled.
-    pub fn requests_stream(
+    /// Given an input stream of source fetches, returns an output stream of fetch results.
+    ///
+    /// Spawns up to `concurrent` + `1` number of concurrent async tasks on the runtime.
+    /// One task for managing the fetch tasks, and one task per fetch request.
+    pub fn stream_from(
         self: Arc<Self>,
-        inputs: impl Stream<Item = (Source, Arc<Data>)> + Unpin + Send + 'static,
-    ) -> impl Stream<
-        Item = impl Future<Output = (Arc<Path>, Arc<Data>, Result<(), Error>)> + Send + 'static,
-    > + Send
-           + Unpin
-           + 'static {
-        inputs.map(move |(source, extra)| {
-            let fetcher = self.clone();
-
-            async move {
-                let Source { dest, urls, part } = source;
-
-                let result = match part {
-                    Some(part) => match fetcher
-                        .clone()
-                        .request(urls, part.clone(), extra.clone())
-                        .await
-                    {
-                        Ok(()) => fs::rename(&*part, &*dest).await.map_err(Error::Rename),
-                        Err(why) => Err(why),
-                    },
-                    None => {
-                        fetcher
-                            .clone()
-                            .request(urls, dest.clone(), extra.clone())
-                            .await
+        inputs: impl Stream<Item = (Source, Arc<Data>)> + Send + 'static,
+        concurrent: usize,
+    ) -> Pin<Box<dyn Stream<Item = AsyncFetchOutput<Data>> + Send + 'static>> {
+        let shutdown = self.shutdown.clone();
+        let cancel_trigger = shutdown.wait_shutdown_triggered();
+        // Takes input requests and converts them into a stream of fetch requests.
+        let stream = inputs
+            .map(move |(Source { dest, urls, part }, extra)| {
+                let fetcher = self.clone();
+                async move {
+                    if fetcher.delay_between_requests != 0 {
+                        let delay = Duration::from_millis(fetcher.delay_between_requests);
+                        tokio::time::sleep(delay).await;
                     }
-                };
 
-                (dest, extra, result)
-            }
-        })
+                    tokio::spawn(async move {
+                        let _token = match fetcher.shutdown.delay_shutdown_token() {
+                            Ok(token) => token,
+                            Err(_) => return (dest, extra, Err(Error::Canceled)),
+                        };
+
+                        let task = async {
+                            match part {
+                                Some(part) => {
+                                    match fetcher.request(urls, part.clone(), extra.clone()).await {
+                                        Ok(()) => {
+                                            fs::rename(&*part, &*dest).await.map_err(Error::Rename)
+                                        }
+                                        Err(why) => Err(why),
+                                    }
+                                }
+                                None => fetcher.request(urls, dest.clone(), extra.clone()).await,
+                            }
+                        };
+
+                        let result = task.await;
+
+                        (dest, extra, result)
+                    })
+                    .await
+                    .unwrap()
+                }
+            })
+            .buffer_unordered(concurrent)
+            .take_until(cancel_trigger);
+
+        Box::pin(stream)
     }
 
     /// Request a file from one or more URIs.
@@ -261,48 +296,77 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
 
         remove_parts(&to).await;
 
+        let attempts = Arc::new(AtomicU16::new(0));
+
         let fetch = || async {
-            self.clone()
-                .inner_request(uris.clone(), to.clone(), extra.clone())
-                .await
-        };
+            loop {
+                let task = self.clone().inner_request(
+                    uris.clone(),
+                    to.clone(),
+                    extra.clone(),
+                    attempts.clone(),
+                );
 
-        let cleanup = || async {
-            remove_parts(&to).await;
+                let result = task.await;
 
-            self.send(|| (to.clone(), extra.clone(), FetchEvent::Fetched));
-        };
+                if let Err(Error::NetworkChanged) | Err(Error::TimedOut) = result {
+                    let mut attempts = 5;
+                    while attempts != 0 {
+                        tokio::time::sleep(Duration::from_secs(3)).await;
 
-        let mut attempts = 0;
+                        let net_check = crate::utils::run_timed(
+                            Some(Duration::from_secs(3)),
+                            crate::utils::network_interrupt(head(&self.client, &uris[0])),
+                        );
 
-        let result = match fetch().await {
-            Ok(()) => Ok(()),
-            Err(mut why) => {
-                while attempts < self.retries {
-                    attempts += 1;
+                        if net_check.await.is_ok() {
+                            tokio::time::sleep(Duration::from_secs(3)).await;
+                            break;
+                        }
 
-                    if self.cancelled() {
-                        cleanup().await;
-                        return Err(Error::Cancelled);
+                        attempts -= 1;
                     }
 
                     self.send(|| (to.clone(), extra.clone(), FetchEvent::Retrying));
+                    remove_parts(&to).await;
+                    tokio::time::sleep(Duration::from_secs(3)).await;
 
-                    match fetch().await {
-                        Ok(()) => {
-                            cleanup().await;
-                            return Ok(());
-                        }
-
-                        Err(cause) => why = cause,
-                    }
+                    continue;
                 }
 
-                Err(why)
+                return result;
             }
         };
 
-        cleanup().await;
+        let task = async {
+            if let Err(mut why) = fetch().await {
+                if let Error::Canceled = why {
+                    return Err(why);
+                }
+
+                while attempts.fetch_add(1, Ordering::SeqCst) < self.retries {
+                    remove_parts(&to).await;
+                    self.send(|| (to.clone(), extra.clone(), FetchEvent::Retrying));
+                    tokio::time::sleep(Duration::from_secs(3)).await;
+
+                    if let Err(source) = fetch().await {
+                        why = source;
+                    }
+                }
+
+                return Err(why);
+            };
+
+            Ok(())
+        };
+
+        let result = task.await;
+
+        remove_parts(&to).await;
+
+        if result.is_ok() {
+            self.send(|| (to.clone(), extra.clone(), FetchEvent::Fetched));
+        }
 
         result
     }
@@ -312,6 +376,7 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
         uris: Arc<[Box<str>]>,
         to: Arc<Path>,
         extra: Arc<Data>,
+        attempts: Arc<AtomicU16>,
     ) -> Result<(), Error> {
         let mut length = None;
         let mut modified = None;
@@ -336,6 +401,7 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
 
                         if metadata.len() == length {
                             if ts.as_secs() == date_as_timestamp(last_modified) {
+                                info!("already fetched {}", to.display());
                                 self.send(|| (to, extra.clone(), FetchEvent::AlreadyFetched));
                                 return Ok(());
                             } else {
@@ -368,7 +434,7 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
                         self.send(|| (to.clone(), extra.clone(), FetchEvent::Progress(resume)));
                     }
 
-                    get_many(
+                    let result = get_many(
                         self.clone(),
                         to.clone(),
                         uris,
@@ -376,8 +442,13 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
                         length,
                         modified,
                         extra,
+                        attempts.clone(),
                     )
-                    .await?;
+                    .await;
+
+                    if let Err(why) = result {
+                        return Err(why);
+                    }
 
                     if let Some(modified) = modified {
                         update_modified(&to, modified)?;
@@ -396,11 +467,12 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
             }
         }
 
-        let mut request = Request::get(&*uris[0]);
+        let mut request = self.client.get(&*uris[0]);
 
         if resume != 0 {
             if let Ok(true) = supports_range(&self.client, &*uris[0], resume, length).await {
                 request = request.header("Range", range::to_string(resume, length));
+                self.send(|| (to.clone(), extra.clone(), FetchEvent::Progress(resume)));
             } else {
                 resume = 0;
             }
@@ -409,29 +481,30 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
         let path = match crate::get(
             self.clone(),
             request,
-            FetchLocation::create(to.clone(), length, resume != 0).await?,
+            FetchLocation::create(to.clone(), resume != 0).await?,
             to.clone(),
-            &mut modified,
             extra.clone(),
+            attempts.clone(),
         )
         .await
         {
-            Ok(path) => path,
-
+            Ok((path, _)) => path,
             Err(Error::Status(StatusCode::NOT_MODIFIED)) => to,
 
             // Server does not support if-modified-since
             Err(Error::Status(StatusCode::NOT_IMPLEMENTED)) => {
-                let request = Request::get(&*uris[0]);
-                crate::get(
+                let request = self.client.get(&*uris[0]);
+                let (path, _) = crate::get(
                     self.clone(),
                     request,
-                    FetchLocation::create(to.clone(), length, resume != 0).await?,
+                    FetchLocation::create(to.clone(), resume != 0).await?,
                     to.clone(),
-                    &mut modified,
                     extra.clone(),
+                    attempts,
                 )
-                .await?
+                .await?;
+
+                path
             }
 
             Err(why) => return Err(why),
@@ -444,12 +517,6 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
         Ok(())
     }
 
-    fn cancelled(&self) -> bool {
-        self.cancel
-            .as_ref()
-            .map_or(false, |cancel| cancel.load(Ordering::SeqCst))
-    }
-
     fn send(&self, event: impl FnOnce() -> (Arc<Path>, Arc<Data>, FetchEvent)) {
         if let Some(sender) = self.events.as_ref() {
             let _ = sender.send(event());
@@ -457,10 +524,10 @@ impl<Data: Send + Sync + 'static> Fetcher<Data> {
     }
 }
 
-pub async fn head(client: &Client, uri: &str) -> Result<Option<Response<AsyncBody>>, Error> {
-    let request = Request::head(uri).body(()).unwrap();
+async fn head(client: &Client, uri: &str) -> Result<Option<Response>, Error> {
+    let request = client.get(uri).build().unwrap();
 
-    match validate(client.send_async(request).await?).map(Some) {
+    match validate(client.execute(request).await?).map(Some) {
         result @ Ok(_) => result,
         Err(Error::Status(StatusCode::NOT_MODIFIED))
         | Err(Error::Status(StatusCode::NOT_IMPLEMENTED)) => Ok(None),
@@ -468,18 +535,19 @@ pub async fn head(client: &Client, uri: &str) -> Result<Option<Response<AsyncBod
     }
 }
 
-pub async fn supports_range(
+async fn supports_range(
     client: &Client,
     uri: &str,
     resume: u64,
     length: Option<u64>,
 ) -> Result<bool, Error> {
-    let request = Request::head(uri)
+    let request = client
+        .head(uri)
         .header("Range", range::to_string(resume, length).as_str())
-        .body(())
+        .build()
         .unwrap();
 
-    let response = client.send_async(request).await?;
+    let response = client.execute(request).await?;
 
     if response.status() == StatusCode::PARTIAL_CONTENT {
         if let Some(header) = response.headers().get("Content-Range") {
@@ -496,27 +564,7 @@ pub async fn supports_range(
     }
 }
 
-pub async fn timed<F, T>(duration: Duration, future: F) -> Result<T, Error>
-where
-    F: Future<Output = T> + Unpin,
-{
-    let timeout = async move {
-        tokio::time::sleep(duration).await;
-        Err(Error::TimedOut)
-    };
-
-    let result = async move { Ok(future.await) };
-
-    futures::pin_mut!(timeout);
-    futures::pin_mut!(result);
-
-    futures::future::select(timeout, result)
-        .await
-        .factor_first()
-        .0
-}
-
-fn validate(response: Response<AsyncBody>) -> Result<Response<AsyncBody>, Error> {
+fn validate(response: Response) -> Result<Response, Error> {
     let status = response.status();
 
     if status.is_informational() || status.is_success() {
@@ -531,7 +579,7 @@ trait ResponseExt {
     fn last_modified(&self) -> Option<HttpDate>;
 }
 
-impl ResponseExt for Response<AsyncBody> {
+impl ResponseExt for Response {
     fn content_length(&self) -> Option<u64> {
         let header = self.headers().get("content-length")?;
         header.to_str().ok()?.parse::<u64>().ok()
